##############################################################################
# Description
#     This is the virtual controller of the machine tool driver for 
#     a generic 5-axis table-table configuration with A and C controlled
#     axis.
#
#     If you use this file as a base for a new driver, fopr start,  look at the 
#     lines that starts with #>>> and modify them accordingly.
#
# Revisions
#
#   Date        Who            Reason
# 01-Jan-2002   yxd            First Version
# 28-Feb-2002   yxd            Added  hole cycles simulation
#
#
# $HISTORY$
#
###############################################################################


#______________________________________________________________________________
# The following command invokes the corresponding event handler to generate the 
# NC program
#______________________________________________________________________________

set the_cam_post_dir  [MOM_ask_env_var UGII_CAM_POST_DIR]
set the_cam_debug_dir [MOM_ask_env_var UGII_CAM_DEBUG_DIR]


# Simulation include files
  source ${the_cam_post_dir}/sim_high_level_sv_commands.tcl

  
#______________________________________________________________________________
# The following command invokes the debugging mode.
#______________________________________________________________________________
#  source ${the_cam_debug_dir}/mom_debug.tcl
#  source ${the_cam_debug_dir}/mom_review_mtd.tcl
#********Setting the debug mode ON/OFF
#   MOM_set_debug_mode OFF


#
# ====== define some local parameters =========
#
  set sim_prev_tool_name "undefined"
  set sim_prev_zcs       "undefined"
  set sim_feedrate_mode  "RAPID"
  set sim_mtd_initialized 0

#****************************************************

#
#========== SET globals for machine tool driver ==============
#
#
#
# As part of the initialization of the machine tool driver, 
# the MOM_SIM_initialize_mtd event is triggered and the MOM_SIM_initialize_mtd 
# procedure is called. First thing that this procedure does it invokes 
# the S&V command to initialize the machine tool driver and to set the virtual 
# controller mode. The S&V command is SIM_mtd_init and it accepts one argument, 
# the controller mode. Currently there are three virtual controller modes:
# 
#    NC_CONTROLLER_SIM_EVENT_HANDLER: A dedicated SIM Event is triggered for each Post
#                                  event. The MTD file (i.e. the TCL file) should
#                                  contains TCL procedures to handle those SIM events.
#                                  For example MOM_SIM_linear_move simulates the linear 
#                                  motion event. Those SIM event handlers should call
#                                  the corresponding S&V commands to move axis, mount 
#                                  tools and devices, set parameters and issue different 
#                                  feedback 
#
#    NC_CONTROLLER_PLUGIN: At each event (when applicable)  MOM POST sets the TCL
#                       global variable MOM_SIM_NC_COMMAND with the next NC program line
#                       to execute. Then the event MOM_SIM_execute_nc_command 
#                       is triggered to process the NC program command stored in variable.
#
#    NC_CONTROLLER_MTD_EVENT_HANDLER: The simulation handling is mixed with NC-program
#                                  handling. No Special handling for MTD is defined.
#                                  S&V commands can be mixed with other post commands.
#
# When MTD ends its execution the event MOM_SIM_exit_mtd is triggered to do the proper 
# cleanup of the machine tool driver.
#
# This files contain an example for setting the machine tool driver to work
# in NC_CONTROLLER_SIM_EVENT_HANDLER. Look at the mill3x_mtd_plugin.tcl for 
# an example on how to set the MTD to work in NC_CONTROLLER_PLUGIN mode.
#
#
#_______________________________________________________________________________
proc MOM_SIM_initialize_mtd { } {
#_______________________________________________________________________________
# This procedure is executed before the simulation starts. it suppose to initialize 
# the machine tool driver to work in SIM event handler mode.
#_______________________________________________________________________________
#
   global sound_files_library sim_mtd_initialized

#---- initializing MTD and setting controller mode. see description of modes above
   SIM_mtd_init NC_CONTROLLER_MTD_EVENT_HANDLER

#
# define the axes mapping between the logical axes known as 
# X Y Z A B C and thier phisical axes as we named them in the
# specific machine tool kinematics model. The first
# letter is for the logical axes and second for phisical
#
# for example
#   set mom_sim_trans_mapping [list set X x-phisical Y y-phisical Z z-phisical]
# 
#
##>>>>>
  SIM_set_linear_axes_map  [concat X X Y Y Z Z]
  SIM_set_rotary_axes_map  [concat B B C C]

#
# For this machine all axes participate in executing a certain motion. 
# so include all axes in axes configuration. Use logical names for axes
##>>>>>
  SIM_set_linear_axes_config  [concat X Y Z]
  SIM_set_rotary_axes_config  [concat B C]

  #set default reference junction
  VNC_set_ref_jct "undefined"

#
# set the callback function to compute duration
  SIM_set_duration_callback_fct "CalculateDurationTime"

#
# Set S&V command degugging. 
# First initialize debugging
# SIM_dbg_start "============== Start Debugging for Driver M5ACTT ========"

#
# Write message to output file
# SIM_dbg_write_message ">>> This is a testing message"

# Set what outputs we want to see. Valid tokens are:
# SIM_DBG_OUTPUT_ERROR, SIM_DBG_OUTPUT_NC_CMD, SIM_DBG_OUTPUT_CEVENT, 
# SIM_DBG_OUTPUT_INPUT_SV_EVENT, SIM_DBG_OUTPUT_OUTPUT_SV_EVENT, 
# SIM_DBG_OUTPUT_PROCESSED_SV_EVENT, SIM_DBG_OUTPUT_SV_CMD, SIM_DBG_OUTPUT_ALL
# 
#  SIM_dbg_set_output ON SIM_DBG_OUTPUT_INPUT_SV_EVENT 

#>>>
# use the following commands to customize the simulation dialog. See Custom
# Dialog Spec for command parameters.
#

# SIM_dialog_start "Control Panel for MILL3AX Driver"

# system items
#SIM_dialog_add_item    "SYS_ACTIVE_TOOL_ITEM    S1"
#SIM_dialog_add_item    "SYS_SPEED_ITEM          S2" 
#SIM_dialog_add_item    "SYS_FEED_ITEM           S3"
#SIM_dialog_add_item    "SYS_COOLANT_ITEM        S4 ORIENTATION=<HORIZONTAL>"
#SIM_dialog_add_item    "SYS_POSITION_ITEM       S5"
#SIM_dialog_add_item    "SYS_PROGRAM_WINDOW_ITEM S6"
#SIM_dialog_add_item    "SYS_MESSAGE_WINDOW_ITEM S7"
#SIM_dialog_add_item    "SYS_STEP_CONTROL_ITEM   S8"
#SIM_dialog_add_item    "SYS_ANIMATION_SPEED_ITEM S9"
#SIM_dialog_add_item    "SYS_MACHINE_TIME_ITEM   S10"

#user items
#SIM_dialog_add_item     "USR_SEPARATOR_ITEM      S11"
#SIM_dialog_add_item     "USR_LABEL_ITEM          S12 LABEL=<my label>"
#SIM_dialog_add_item     "USR_STRING_ITEM         S13 LABEL=<my string> VALUE=<string value>"
#SIM_dialog_add_item     "USR_RADIO_ITEM          S14 LABEL=<my radio> ORIENTATION=<HORIZONTAL> SPAN=<3> LIST=<radio1,radio2,radio3> FORMAT=<CHECK_BOX>"

# SIM_dialog_end 

# for internal use flag to check whther the post is running in simulation
# or for NC files creation.
  set sim_mtd_initialized 1
}

#_______________________________________________________________________________
proc MOM_SIM_exit_mtd { } {
#_______________________________________________________________________________
# This procedure exits the machine tool driver. this procedure 
# is called when the machine tool simulation ends.
#_______________________________________________________________________________
#

# invoke S&V command to exit machine tool driver. 
     SIM_mtd_reset 

# end debugging
#   SIM_dbg_end 
}

#=============================================================
proc VNC_start_of_program { } {
#=============================================================
#_______________________________________________________________________________
# This procedure is executed at the very begining of the program.
# It gets called before any command is read from the task file.
# here we can do any machine tool initialization 
#_______________________________________________________________________________

  global sound_files_library 
  global mom_output_unit
  global sim_mtd_initialized

  # check if simulation has been initialized properly
  if { !$sim_mtd_initialized } { return }

  SIM_start_of_simulation

# set the machine tool driver units
  if { $mom_output_unit == "MM" } {
     SIM_set_mtd_units "MM" 
  } else {
     SIM_set_mtd_units "INCH" 
  }

 #
 #define the reference junction as the spindle junction
 VNC_set_ref_jct "undefined"

# setup tool change position and head position
###  VNC_set_tool_change_position

}

#=============================================================
proc VNC_end_of_program { } {
#=============================================================
    global sound_files_library 
    global sim_prev_tool_name
    global sim_mtd_initialized
    global sim_4th_axis sim_5th_axis

  # check if simulation has been initialized properly
  if { !$sim_mtd_initialized } { return }
   
     VNC_unmount_tool $sim_prev_tool_name

     SIM_move_linear_zcs $sim_4th_axis 0 $sim_5th_axis 0

    SIM_end_of_simulation
}


#
# create and set the current ZCS junction
# this function is called when the MOM_msys is triggered
#
#=============================================================
proc VNC_msys { mount_zcs_on } {
#=============================================================
#
# this event is trigered when the msys is changed
#
 global mom_msys_origin mom_msys_matrix 
 global sim_rotate_table_switch  mom_sim_result 
 global sim_mtd_initialized

 # check if simulation has been initialized properly
 if { !$sim_mtd_initialized } { return }

 # mount the junction on the rotary table 
 set dest_cmp $mount_zcs_on

 #
 # convert part/simulation units to MTD units
 lappend  origin [ SIM_convert_sim_to_mtd_units $mom_msys_origin(0) ]
 lappend  origin [ SIM_convert_sim_to_mtd_units $mom_msys_origin(1) ]
 lappend  origin [ SIM_convert_sim_to_mtd_units $mom_msys_origin(2) ]

 # convert matrix vector to list
 for {set i 0} {$i<9} {incr i}  {
   lappend mtx $mom_msys_matrix($i)
 }

 # check if ZCS junction already created. if yes, return its name. if not create new one
 # and return its name. this optimization is needed because this event is invoked
 # before each operation is exceuted.
 set zcs_name [ SIM_create_zcs_junction CURRENT_ZCS_JCT $dest_cmp $origin  $mtx ]
 
 SIM_update_current_zcs $zcs_name
}

#=============================================================
proc VNC_send_message { message } {
#=============================================================
 global sim_mtd_initialized

 # check if simulation has been initialized properly
 if { !$sim_mtd_initialized } { return }


  SIM_feedback_message "MTD:" "$message"
  
}

#-------------------------------------------------------------------------------
# CalculateDurationTime
#
# return the duration time in seconds to pass the delta length in
# current feedrate
# linear_or_angular is either "LINEAR" or "ANGULAR" which defines
# whether delta is linear(in current units) or angular(in degrees) 
# displacement 
#-------------------------------------------------------------------------------
#=============================================================
proc CalculateDurationTime { linear_or_angular delta } {
#=============================================================

	global mom_feed_rate_mode
	global mom_feed_rate
	global mom_output_unit sim_feedrate_mode

    # for now we don't compute durations for angular motions
    if { $linear_or_angular == "ANGULAR" } { return 0}

    if { $sim_feedrate_mode == "RAPID" } {
      # for now just set an arbitrary RAPID feedrate. this has to match
      # the controller specifications
      # 10 meters per minute as rapid speed

      set feed 10000 
    } else {
	  if {$mom_feed_rate_mode != "MMPM" } \
		{ set feed [expr {$mom_feed_rate * 25.4}] } \
	  else \
		{ set feed $mom_feed_rate }
    }

    set length [expr {abs($delta)}]

	if {$mom_output_unit != "MM" } \
		{ set length [expr {$length * 25.4}] }

	
	# if [catch{ set ret [expr ($length / $feed) * 60.0]}] {return 0.0}
	if {[expr {abs($feed)}] < 0.00001 || $length < 0.00001 } \
		{ set time 0 } \
	else \
		{ set time [expr {($length / $feed) * 60.0}] }

    return $time
}


#=============================================================
proc VNC_circular_move { } {
#=============================================================

  global mom_feed_rate
  global mom_spindle_speed
  global mom_arc_direction
  global mom_pos_arc_center
  global mom_pos_arc_axis
  global mom_pos mom_out_angle_pos
  global mom_arc_radius
  global mom_sim_zero_tol
  global sim_mtd_initialized

  # check if simulation has been initialized properly
  if { !$sim_mtd_initialized } { return }

  VNC_set_feedrate_mode "CUT"
  SIM_set_speed $mom_spindle_speed [VNC_ask_speed_mode]

  if {$mom_arc_direction == "CLW"} {
     set dir -1.0
  } else {
     set dir 1.0
  }

  set i $mom_pos_arc_center(0) 
  set j $mom_pos_arc_center(1)
  set k $mom_pos_arc_center(2)
  
# Store original value of mom_sim_zero_tol and reset tol to compensate for circle center errors

  set temp_mom_sim_zero_tol $mom_sim_zero_tol
  set mom_sim_zero_tol 0.005

  if {[EQ_is_equal abs($mom_pos_arc_axis(2)) 1.0]} {
#     SIM_move_circular_zcs 0 0 $dir X $mom_pos(0) Y $mom_pos(1) R $mom_arc_radius
     SIM_move_circular_zcs 0 0 $dir X $mom_pos(0) Y $mom_pos(1) I $i J $j
  } elseif {[EQ_is_equal abs($mom_pos_arc_axis(1)) 1.0]} {
#     SIM_move_circular_zcs 0 $dir 0 Z $mom_pos(2) X $mom_pos(0) R $mom_arc_radius
     SIM_move_circular_zcs 0 $dir 0 Z $mom_pos(2) X $mom_pos(0) K $k I $i
  } elseif {[EQ_is_equal abs($mom_pos_arc_axis(0)) 1.0]} {
#     SIM_move_circular_zcs $dir 0 0 Y $mom_pos(1) Z $mom_pos(2) R $mom_arc_radius
     SIM_move_circular_zcs $dir 0 0 Y $mom_pos(1) Z $mom_pos(2) J $j K $k
  }
# Reset mom_sim_zero_tol to original value
  set mom_sim_zero_tol $temp_mom_sim_zero_tol

}

proc VNC_ask_feedrate_mode { } {
#_______________________________________________________________________________
# This function returns the feedrate units as expected by S&V commands
#_______________________________________________________________________________
  global   mom_feed_rate_mode
  global sim_mtd_initialized

  # check if simulation has been initialized properly
  if { !$sim_mtd_initialized } { return }

  if {[hiset mom_feed_rate_mode]} {
   switch $mom_feed_rate_mode {
    IPM     { set feed_units "INCH_PER_MIN" }
    MMPM    { set feed_units "MM_PER_MIN"}
    IPR     { set feed_units "INCH_PER_REV" }
    MMPR    { set feed_units "MM_PER_REV" }
#    FRN    { set feed_units "MM_PER_100REV" }
    default { set feed_units "UNKNOWN FEEDRATE UNITS" }
   }
  } else {
    set feed_units "UNKNOWN FEEDRATE UNITS" 
  }

  return $feed_units 

}

proc VNC_ask_speed_mode { } {
#_______________________________________________________________________________
# This function returns the speed units as expected by S&V commands
#_______________________________________________________________________________
  global   mom_spindle_mode
  global sim_mtd_initialized

  # check if simulation has been initialized properly
  if { !$sim_mtd_initialized } { return }

  if {[hiset mom_spindle_mode]} {
   switch $mom_spindle_mode {
    RPM { set speed_units "REV_PER_MIN" }
#    SFM {  }
#    SMM {  }
    default { set speed_units "UNKNOWN SPEED UNITS" }
   }
  } else {
    set speed_units "UNKNOWN SPEED UNITS" 
  }

  return $speed_units
}


#=============================================================
proc VNC_set_feedrate_mode { cutting_mode } {
#=============================================================
  global mom_feed_rate sim_feedrate_mode
  global sim_mtd_initialized

  # check if simulation has been initialized properly
  if { !$sim_mtd_initialized } { return }

  SIM_set_cutting_mode $cutting_mode

  set sim_feedrate_mode $cutting_mode

  if { $cutting_mode == "RAPID" } {
    # we need here the actual rapid speed. meanwhile we use what we have
    SIM_set_feed $mom_feed_rate [VNC_ask_feedrate_mode]
  } else {
    SIM_set_feed $mom_feed_rate [VNC_ask_feedrate_mode]
  }
}

#=============================================================
proc VNC_unmount_tool { tool_ug_name } {
#=============================================================

    global sim_prev_tool_name mom_tool_number
    global mom_sim_result sim_tool_change_pos mom_sys_home_pos sim_previous_tool_number
    
    global sim_mtd_initialized

    # check if simulation has been initialized properly
    if { !$sim_mtd_initialized } { return }

       VNC_set_feedrate_mode "RAPID"

       #
       #define the reference junction so the tool change position will
       # be i.r.t machine zero
       VNC_set_ref_jct "undefined"

#>>> move to tool change position... identify the right one

#	if { $sim_previous_tool_number != 0 } \
#	{
#          set tool_mag_index 1
#          if {$sim_previous_tool_number < 47 } { set tool_mag_index 0 }
#          SIM_move_linear_mtcs  Z $sim_tool_change_pos($tool_mag_index,2)   
#          SIM_move_linear_mtcs  X $sim_tool_change_pos($tool_mag_index,0) Y $sim_tool_change_pos($tool_mag_index,1)   
#          SIM_move_linear_mtcs  Z $sim_tool_change_pos($tool_mag_index,5)   
#          SIM_move_linear_mtcs  X $sim_tool_change_pos($tool_mag_index,3) Y $sim_tool_change_pos($tool_mag_index,4)   
#        }
          SIM_move_linear_mtcs  Z $sim_tool_change_pos(2)   
          SIM_move_linear_mtcs  X $sim_tool_change_pos(0) Y $sim_tool_change_pos(1)   
        

       # move to tool change position
#        SIM_move_linear_mtcs  Z $mom_sys_home_pos(2)   
#       SIM_move_linear_mtcs  Y $sim_tool_change_pos(1)
#       SIM_move_linear_mtcs  X $sim_tool_change_pos(0)

       if {$tool_ug_name  != "undefined" } { 
         SIM_ask_kim_comp_name_by_id "TOOL" $tool_ug_name
         set t_name $mom_sim_result
         
         SIM_unmount_tool "2.0" $t_name  
         SIM_update

         set sim_prev_tool_name "undefined"
       }
#	if { $sim_previous_tool_number != 0 } \
#	{
#          set tool_mag_index 1
#          if {$sim_previous_tool_number < 47 } { set tool_mag_index 0 }
#          SIM_move_linear_mtcs  X $sim_tool_change_pos($tool_mag_index,0) Y $sim_tool_change_pos($tool_mag_index,1)   
#	}
}

#=============================================================
proc VNC_tool_change { tool_ug_name } {
#=============================================================

    global sim_prev_tool_name 
    global sound_files_library
    global mom_tool_z_offset mom_tool_number sim_tool_change_pos sim_tool_measure_pos sim_previous_tool_number
    global mom_sim_result sim_current_tool
 
    global sim_mtd_initialized

    # check if simulation has been initialized properly
    if { !$sim_mtd_initialized } { return }

       set tool_name " "

       if {$sim_prev_tool_name != $tool_ug_name && $tool_ug_name  != ""} {

         VNC_unmount_tool $sim_prev_tool_name

# GOTO load tool position based on tool number

        SIM_move_linear_mtcs  Z $sim_tool_change_pos(2)   
        SIM_move_linear_mtcs  X $sim_tool_change_pos(0) Y $sim_tool_change_pos(1)   

#>>> Specify where (i.e. on which K-Component) the tool should be mounted 
#    (e.g SPINDLE) and what junction on that component should be used for mounting
#	(e.g, SPINDLE_JCT).

         SIM_mount_tool "15" "UG_NAME" $tool_ug_name "SPINDLE" "SPINDLE_JCT"
         set tool_name $mom_sim_result 
         set sim_current_tool $tool_name

         if {$tool_name != ""} {
            SIM_activate_tool $tool_name
         }

         SIM_update 

         set sim_prev_tool_name $tool_ug_name
       }

        #
        # set the reference junction
        VNC_set_ref_jct $tool_name

	# output tool change and measure movements

#        SIM_move_linear_mtcs  X $sim_tool_measure_pos(0) Y $sim_tool_measure_pos(1)   
#        SIM_move_linear_mtcs  Z $sim_tool_measure_pos(2)   
#        SIM_move_linear_mtcs  X [expr $sim_tool_measure_pos(0)+41]
#        SIM_move_linear_mtcs  Z [expr $sim_tool_measure_pos(2)-5]   
#        VNC_set_ref_jct "undefined"
#        SIM_move_linear_mtcs  Z $sim_tool_change_pos($tool_mag_index,2)   
#        VNC_set_ref_jct $tool_name
#        
        set sim_previous_tool_number $mom_tool_number

}


#=============================================================
proc VNC_set_ref_jct { tool_name } {
#=============================================================
global sim_mtd_initialized

# check if simulation has been initialized properly
if { !$sim_mtd_initialized } { return }

         if {$tool_name != " " && $tool_name != "" && $tool_name != "undefined"} \
         {
              set ref_jct $tool_name@SIM_TOOL_TIP
         } else { 

#>>>>> when the tool is not mounted, the tool tip is not the reference 
# junction any more. Then identify a new junction. Normaly this should be the
# junction the defines the pivot point.

              #the pivot junction   
              set ref_jct SPINDLE_JCT
         } 

         SIM_set_current_ref_junction $ref_jct  
}

#=============================================================
proc VNC_coolant_on { } {
#=============================================================
  global  sound_files_library sim_coolant_status  
  global sim_mtd_initialized

  # check if simulation has been initialized properly
  if { !$sim_mtd_initialized } { return }

  set sim_coolant_status "ON"

  SIM_set_coolant "ON"
#  SIM_update
}


#=============================================================
proc VNC_coolant_off { } {
#=============================================================

  global  sound_files_library sim_coolant_status 
  global sim_mtd_initialized

  # check if simulation has been initialized properly
  if { !$sim_mtd_initialized } { return }

  set sim_coolant_status "OFF"

  SIM_set_coolant "OFF"
#  SIM_update
}

#=============================================================
proc VNC_move_linear_mtcs { mode args } {
#=============================================================
  global sim_mtd_initialized

  # check if simulation has been initialized properly
  if { !$sim_mtd_initialized } { return }

# remove outer braces 
  set args [join $args]


  if { $mode != "CURRENT" } {
      VNC_set_feedrate_mode $mode
  }

  eval SIM_move_linear_mtcs $args

}

#=============================================================
proc VNC_move_linear_zcs { mode args } {
#=============================================================
  global sim_mtd_initialized

  # check if simulation has been initialized properly
  if { !$sim_mtd_initialized } { return }

# remove outer braces 
  set args [join $args]


  if { $mode != "CURRENT" } {
      VNC_set_feedrate_mode $mode
  }

  eval SIM_move_linear_zcs $args
  
}

#=============================================================
proc VNC_rapid_bc { } {
#=============================================================
  global sim_mtd_initialized 
  global B C
  global gkw_B_work_plane gkw_C_work_plane
  global sim_4th_axis sim_5th_axis

  # check if simulation has been initialized properly
  if { !$sim_mtd_initialized } { return }

  VNC_set_feedrate_mode "RAPID"
  VNC_linear_lintol B C
#  SIM_move_linear_zcs    $sim_4th_axis [expr $B + $gkw_B_work_plane]  $sim_5th_axis [expr $C + $gkw_C_work_plane]

}

#=============================================================
proc VNC_move_rapid_bc {Bin Cin} {
#=============================================================
  global sim_mtd_initialized 
  global sim_4th_axis sim_5th_axis
  global B C mom_pos
  global gkw_B_work_plane gkw_C_work_plane
	global B_sim_old_pos C_sim_old_pos
  
  # check if simulation has been initialized properly
  if { !$sim_mtd_initialized } { return }

  set gkw_B_work_plane $Bin
  set gkw_C_work_plane $Cin
  
  VNC_set_feedrate_mode "RAPID"
  VNC_linear_lintol ZCS B C
#  SIM_move_linear_zcs    $sim_4th_axis $Bin  $sim_5th_axis $Cin

  set B_sim_old_pos $Bin
  set C_sim_old_pos $Cin

}

#=============================================================
proc VNC_rapid_xy { } {
#=============================================================
  global sim_mtd_initialized mom_pos mom_spindle_speed
  global sim_4th_axis sim_5th_axis
  global gkw_B_work_plane gkw_C_work_plane
  global B C

  # check if simulation has been initialized properly
  if { !$sim_mtd_initialized } { return }

  SIM_set_speed $mom_spindle_speed [VNC_ask_speed_mode]
  VNC_set_feedrate_mode "RAPID"
  VNC_linear_lintol X Y Z B C
#  SIM_move_linear_zcs   X $mom_pos(0) Y $mom_pos(1) $sim_4th_axis [expr $B + $gkw_B_work_plane]  $sim_5th_axis [expr $C + $gkw_C_work_plane]

}

#=============================================================
proc VNC_cycle_rapid_z { } {
#=============================================================
  global sim_mtd_initialized mom_spindle_speed mom_cycle_retract_to_pos

  # check if simulation has been initialized properly
  if { !$sim_mtd_initialized } { return }

  SIM_set_speed $mom_spindle_speed [VNC_ask_speed_mode]
  VNC_set_feedrate_mode "RAPID"
  VNC_linear_lintol Z
#  SIM_move_linear_zcs   Z $mom_cycle_retract_to_pos(2) 
}

#=============================================================
proc VNC_rapid_z { } {
#=============================================================
  global sim_mtd_initialized mom_pos mom_spindle_speed

  # check if simulation has been initialized properly
  if { !$sim_mtd_initialized } { return }

  SIM_set_speed $mom_spindle_speed [VNC_ask_speed_mode]
  VNC_set_feedrate_mode "RAPID"
  VNC_linear_lintol Z
#  SIM_move_linear_zcs   Z $mom_pos(2) 
}

#=============================================================
proc VNC_rapid_xyz { } {
#=============================================================
  global sim_mtd_initialized mom_pos mom_spindle_speed
  global sim_4th_axis sim_5th_axis
  global gkw_B_work_plane gkw_C_work_plane
  global B C

  # check if simulation has been initialized properly
  if { !$sim_mtd_initialized } { return }

  VNC_set_feedrate_mode "RAPID"
  SIM_set_speed $mom_spindle_speed [VNC_ask_speed_mode]
  VNC_coolant_on

  VNC_linear_lintol X Y Z B C
#  SIM_move_linear_zcs   X $mom_pos(0) Y $mom_pos(1) Z $mom_pos(2) $sim_4th_axis [expr $B + $gkw_B_work_plane]  $sim_5th_axis [expr $C + $gkw_C_work_plane]

}

#=============================================================
proc VNC_rapid_probe { } {
#=============================================================
  global sim_mtd_initialized sim_tool_change_pos sim_tool_measure_pos mom_spindle_speed
  global sim_current_tool
  global mom_pos

  # check if simulation has been initialized properly
  if { !$sim_mtd_initialized } { return }

  set mom_pos(0) $sim_tool_measure_pos(0)
  set mom_pos(1) $sim_tool_measure_pos(1)
  set mom_pos(2) $sim_tool_measure_pos(2)

  VNC_set_feedrate_mode "RAPID"
  SIM_set_speed 3000 [VNC_ask_speed_mode]
  VNC_coolant_off

  VNC_set_ref_jct "undefined"
  VNC_linear_lintol MTCS X Y
#  SIM_move_linear_mtcs   X $sim_tool_measure_pos(0) Y $sim_tool_measure_pos(1)
  VNC_set_ref_jct $sim_current_tool
  VNC_linear_lintol MTCS Z
#  SIM_move_linear_mtcs   Z $sim_tool_measure_pos(2)
  VNC_set_ref_jct "undefined"
  set mom_pos(2) $sim_tool_change_pos(2)
  VNC_linear_lintol MTCS Z
#  SIM_move_linear_mtcs   Z $sim_tool_change_pos(2)
  VNC_set_ref_jct $sim_current_tool

}


#=============================================================
proc VNC_rapid_home { } {
#=============================================================
  global sim_mtd_initialized gkw_sim_home_pos
  global sim_current_tool
  global mom_pos
  
  # check if simulation has been initialized properly
  if { !$sim_mtd_initialized } { return }

  VNC_set_feedrate_mode "RAPID"

  set mom_pos(0) $gkw_sim_home_pos(0)
  set mom_pos(1) $gkw_sim_home_pos(1)
  set mom_pos(2) $gkw_sim_home_pos(2)

  VNC_set_ref_jct "undefined"
  VNC_linear_lintol MTCS Z
  VNC_linear_lintol MTCS X Y
#  SIM_move_linear_mtcs   Z $gkw_sim_home_pos(2)
#  SIM_move_linear_mtcs   X $gkw_sim_home_pos(0) Y $gkw_sim_home_pos(1)
  VNC_set_ref_jct $sim_current_tool

}


#=============================================================
proc VNC_linear { } {
#=============================================================
  global sim_mtd_initialized mom_pos mom_spindle_speed
  global sim_4th_axis sim_5th_axis
  global gkw_B_work_plane gkw_C_work_plane
  global B C
	global X_sim_pos Y_sim_pos Z_sim_pos B_sim_pos C_sim_pos
	global X_sim_old_pos Y_sim_old_pos Z_sim_old_pos B_sim_old_pos C_sim_old_pos
  global tollin
  # check if simulation has been initialized properly
  if { !$sim_mtd_initialized } { return }

  VNC_set_feedrate_mode "CUT"
  SIM_set_speed $mom_spindle_speed [VNC_ask_speed_mode]
  VNC_coolant_on

  VNC_linear_lintol X Y Z B C
#  SIM_move_linear_zcs   X $mom_pos(0) Y $mom_pos(1) Z $mom_pos(2) $sim_4th_axis [expr $B + $gkw_B_work_plane]  $sim_5th_axis [expr $C + $gkw_C_work_plane]

}

#=============================================================
proc VNC_linear_xy { } {
#=============================================================
  global sim_mtd_initialized mom_pos mom_spindle_speed
  global sim_4th_axis sim_5th_axis
  global gkw_B_work_plane gkw_C_work_plane
  global B C

  # check if simulation has been initialized properly
  if { !$sim_mtd_initialized } { return }

  VNC_set_feedrate_mode "CUT"
  SIM_set_speed $mom_spindle_speed [VNC_ask_speed_mode]
  VNC_coolant_on

  VNC_linear_lintol X Y B C
#  SIM_move_linear_zcs   X $mom_pos(0) Y $mom_pos(1) $sim_4th_axis [expr $B + $gkw_B_work_plane]  $sim_5th_axis [expr $C + $gkw_C_work_plane]

}

#=============================================================
proc VNC_linear_z { } {
#=============================================================
  global sim_mtd_initialized mom_pos mom_spindle_speed
  global sim_4th_axis sim_5th_axis
  global gkw_B_work_plane gkw_C_work_plane
  global B C

  # check if simulation has been initialized properly
  if { !$sim_mtd_initialized } { return }

  VNC_set_feedrate_mode "CUT"
  SIM_set_speed $mom_spindle_speed [VNC_ask_speed_mode]
  VNC_coolant_on

  VNC_linear_lintol Z
#  SIM_move_linear_zcs   Z $mom_pos(2) 

}

#=============================================================
proc VNC_set_post_kinematics { zcs_jct X_axis Y_axis Z_axis {4th_axis ""} {5th_axis ""} {pivot_jct ""} {gauge_jct ""} } {
#=============================================================
  global sim_mtd_initialized

  # check if simulation has been initialized properly
  if { !$sim_mtd_initialized } { return }

  SIM_set_post_kinematics $zcs_jct $X_axis $Y_axis $Z_axis  $4th_axis $5th_axis $pivot_jct $gauge_jct
}

#=============================================================
proc VNC_is_machine_tool_model_exists { } {
#=============================================================
  global sim_mtd_initialized

  # this function should realy check for whether a machine tool
  # model exist or not. for now, we check if simulation has 
  # been initialized properly
  
  return $sim_mtd_initialized  

}

#=============================================================
proc VNC_macro_def { macro_name } {
#=============================================================
  global sim_mtd_initialized

  # check if simulation has been initialized properly
  if { !$sim_mtd_initialized } { return }


  SIM_macro_def $macro_name

}

#=============================================================
proc VNC_macro_end { macro_name } {
#=============================================================
  global sim_mtd_initialized

  # check if simulation has been initialized properly
  if { !$sim_mtd_initialized } { return }


  SIM_macro_end $macro_name

}

#=============================================================
proc VNC_macro_call { macro_name } {
#=============================================================
  global sim_mtd_initialized

  # check if simulation has been initialized properly
  if { !$sim_mtd_initialized } { return }


  SIM_macro_call $macro_name

}

#proc VNC_do_drill_cycle { } {
##=============================================================
#  global sim_mtd_initialized 
#  global mom_cycle_rapid_to_pos 
#  global mom_cycle_feed_to_pos 
#  global mom_out_angle_pos
#  global mom_cycle_spindle_axis
#  global sim_4th_axis sim_5th_axis
#
#  # check if simulation has been initialized properly
#  if { !$sim_mtd_initialized } { return }
#
#  set axis_name(0) X
#  set axis_name(1) Y
#  set axis_name(2) Z
#   
#  VNC_move_linear_zcs CUT \
#       X $mom_cycle_feed_to_pos(0) \
#       Y $mom_cycle_feed_to_pos(1) \
#       Z $mom_cycle_feed_to_pos(2) \
#       $sim_4th_axis $mom_out_angle_pos(0) \
#       $sim_5th_axis $mom_out_angle_pos(1)
#
#  VNC_move_linear_zcs RAPID \
#       $axis_name($mom_cycle_spindle_axis) $mom_cycle_rapid_to_pos($mom_cycle_spindle_axis)
#}
#

#----------------------------------------------------------------------------
#
# Function VNC_reload_kinematics updates the post kinematics variables according to 
# the machine tool kinematics model.
#
#
#----------------------------------------------------------------------------
proc VNC_reload_kinematics {} {

  global sim_prev_zcs sim_4th_axis  sim_5th_axis 

  set uf_library "ugpadvkins.dll"

  # define the reference junction for defining the 4th and 5th
  # axis post kinematics data. It can be the ZCS junction or 
  # an explicit one. Offsets and vectors for kinematics are defined
  # i.r.t this junction. If you use ZCS, make sure that the right 
  # ZCS is active when this procedure is called.
#>>>>

## set current_zcs [SIM_ask_current_zcs_junction]
 set current_zcs ZERO_JCT

  #
  # set kinematics based on the machine tool kinematics model
  # current_zcs represents the MCS used with this machine
  if { [ VNC_is_machine_tool_model_exists ] } {
      if { $sim_prev_zcs != $current_zcs } {
         VNC_set_post_kinematics $current_zcs X Y Z $sim_4th_axis $sim_5th_axis 
         set sim_prev_zcs $current_zcs
      }
  }

##  MOM_run_user_function [MOM_ask_env_var UGII_CAM_AUXILIARY_DIR]$uf_library ufusr
   MOM_reload_kinematics
}

#=============================================================
proc VNC_do_drill_cycle { } {
#=============================================================
  global sim_mtd_initialized 
  global mom_cycle_rapid_to_pos 
  global mom_cycle_feed_to_pos 
  global mom_cycle_retract_to_pos
  global mom_cycle_delay
  global mom_cycle_rapid_to 
  global mom_cycle_feed_to
  global mom_cycle_retract_to
  global mom_out_angle_pos
  global mom_cycle_spindle_axis
  global sim_4th_axis sim_5th_axis

  # check if simulation has been initialized properly
  if { !$sim_mtd_initialized } { return }
   
  #set axis_name(0) X
  #set axis_name(1) Y
  #set axis_name(2) Z

  VNC_set_feedrate_mode "RAPID"
  VNC_move_linear_zcs RAPID \
       X $mom_cycle_rapid_to_pos(0) \
       Y $mom_cycle_rapid_to_pos(1) \
       Z $mom_cycle_rapid_to_pos(2) ;#\
#       $sim_4th_axis $mom_out_angle_pos(0) \
#       $sim_5th_axis $mom_out_angle_pos(1)

  VNC_set_feedrate_mode "CUT"
  VNC_move_linear_zcs CUT \
       X $mom_cycle_feed_to_pos(0) \
       Y $mom_cycle_feed_to_pos(1) \
       Z $mom_cycle_feed_to_pos(2) 

  VNC_set_feedrate_mode "RAPID"
  VNC_move_linear_zcs RAPID \
       X $mom_cycle_retract_to_pos(0) \
       Y $mom_cycle_retract_to_pos(1) \
       Z $mom_cycle_retract_to_pos(2) 

#  VNC_move_linear_zcs RAPID \
#       $axis_name($mom_cycle_spindle_axis) $mom_cycle_rapid_to_pos($mom_cycle_spindle_axis)
}

#=============================================================
proc VNC_do_drill_dwell_cycle { } {
#=============================================================
  global sim_mtd_initialized 
  global mom_cycle_rapid_to_pos 
  global mom_cycle_feed_to_pos 
  global mom_cycle_retract_to_pos
  global mom_cycle_delay
  global mom_out_angle_pos
  global sim_4th_axis sim_5th_axis
  global mom_cycle_rapid_to 
  global mom_cycle_feed_to
  global mom_cycle_retract_to
 
  # check if simulation has been initialized properly
  if { !$sim_mtd_initialized } { return }

  VNC_move_linear_zcs RAPID \
       X $mom_cycle_rapid_to_pos(0) \
       Y $mom_cycle_rapid_to_pos(1) \
       Z $mom_cycle_rapid_to_pos(2) \
       $sim_4th_axis $mom_out_angle_pos(0) \
       $sim_5th_axis $mom_out_angle_pos(1)

  VNC_move_linear_zcs CUT \
       X $mom_cycle_feed_to_pos(0) \
       Y $mom_cycle_feed_to_pos(1) \
       Z $mom_cycle_feed_to_pos(2) 

  SIM_delay $mom_cycle_delay

  VNC_move_linear_zcs RAPID \
       X $mom_cycle_retract_to_pos(0) \
       Y $mom_cycle_retract_to_pos(1) \
       Z $mom_cycle_retract_to_pos(2) 

}

#=============================================================
 proc vnc_update_last_tool_pos_and_axis { } {
#=============================================================
  global sim_last_X sim_last_Y sim_last_Z
  global sim_last_tool_axis_X sim_last_tool_axis_Y sim_last_tool_axis_Z

  # update last position and tool axis needed by s&v commands
  set pos_and_axis [SIM_ask_tool_position_zcs [SIM_ask_current_ref_junction] ]

  set sim_last_X [lindex $pos_and_axis 0] 
  set sim_last_Y [lindex $pos_and_axis 1]   
  set sim_last_Z [lindex $pos_and_axis 2] 

  set sim_last_tool_axis_X [lindex $pos_and_axis 3] 
  set sim_last_tool_axis_Y [lindex $pos_and_axis 4]   
  set sim_last_tool_axis_Z [lindex $pos_and_axis 5] 
}

#=============================================================
 proc vnc_get_tool_pos { dist } {
#=============================================================
  global sim_last_X sim_last_Y sim_last_Z
  global sim_last_tool_axis_X sim_last_tool_axis_Y sim_last_tool_axis_Z

  set x [expr  $sim_last_X + $dist * $sim_last_tool_axis_X]
  set y [expr  $sim_last_Y + $dist * $sim_last_tool_axis_Y]
  set z [expr  $sim_last_Z + $dist * $sim_last_tool_axis_Z]

  return "X $x Y $y Z $z"
}


#=============================================================
proc VNC_do_drill_deep_cycle { } {
#=============================================================
  global sim_mtd_initialized 
  global mom_cycle_rapid_to_pos 
  global mom_cycle_feed_to_pos 
  global mom_cycle_retract_to_pos
  global mom_cycle_rapid_to 
  global mom_cycle_feed_to
  global mom_cycle_retract_to
  global mom_cycle_step1
  global mom_out_angle_pos
  global sim_4th_axis sim_5th_axis

  # check if simulation has been initialized properly
  if { !$sim_mtd_initialized } { return }

  VNC_move_linear_zcs RAPID \
       X $mom_cycle_rapid_to_pos(0) \
       Y $mom_cycle_rapid_to_pos(1) \
       Z $mom_cycle_rapid_to_pos(2) \
       $sim_4th_axis $mom_out_angle_pos(0) \
       $sim_5th_axis $mom_out_angle_pos(1)

  # update the last position 
  vnc_update_last_tool_pos_and_axis

  set abs_step1      [expr abs($mom_cycle_step1)]
  set abs_feed_to    [expr abs($mom_cycle_feed_to)]
  set abs_rapid_to   [expr abs($mom_cycle_rapid_to)]
  set abs_retract_to [expr abs($mom_cycle_retract_to)]

  if { $abs_step1 > 0 && $abs_feed_to > 0 } {

    set dist [expr $abs_step1 + $abs_rapid_to]
    set prev_dist 0

    while { $dist < $abs_feed_to } {

      VNC_move_linear_zcs CUT [ vnc_get_tool_pos [expr -1*$dist] ] 

      if { $prev_dist > 0 } {
        VNC_move_linear_zcs RAPID [ vnc_get_tool_pos [expr -1*$prev_dist]] 
       } else {
        VNC_move_linear_zcs RAPID [ vnc_get_tool_pos [expr -0.90*$dist]] 
	   }


      set prev_dist $dist
      set dist [expr $dist + $abs_step1 ]
    }
  }

  VNC_move_linear_zcs CUT \
       X $mom_cycle_feed_to_pos(0) \
       Y $mom_cycle_feed_to_pos(1) \
       Z $mom_cycle_feed_to_pos(2) 

  VNC_move_linear_zcs RAPID \
       X $mom_cycle_retract_to_pos(0) \
       Y $mom_cycle_retract_to_pos(1) \
       Z $mom_cycle_retract_to_pos(2) 
}

#=============================================================
proc VNC_do_drill_break_chip_cycle { } {
#=============================================================
  global sim_mtd_initialized 
  global mom_cycle_rapid_to_pos 
  global mom_cycle_feed_to_pos 
  global mom_cycle_retract_to_pos
  global mom_cycle_rapid_to 
  global mom_cycle_feed_to
  global mom_cycle_retract_to
  global mom_cycle_step1
  global mom_cycle_step2
  global mom_out_angle_pos
  global sim_4th_axis sim_5th_axis

  # check if simulation has been initialized properly
  if { !$sim_mtd_initialized } { return }

  VNC_move_linear_zcs RAPID \
       X $mom_cycle_rapid_to_pos(0) \
       Y $mom_cycle_rapid_to_pos(1) \
       Z $mom_cycle_rapid_to_pos(2) \
       $sim_4th_axis $mom_out_angle_pos(0) \
       $sim_5th_axis $mom_out_angle_pos(1)

  # update the last position 
  vnc_update_last_tool_pos_and_axis

  set abs_step1      [expr abs($mom_cycle_step1)]
  set abs_step2      [expr abs($mom_cycle_step2)]
  set abs_feed_to    [expr abs($mom_cycle_feed_to)]
  set abs_rapid_to   [expr abs($mom_cycle_rapid_to)]
  set abs_retract_to [expr abs($mom_cycle_retract_to)]

  if { $abs_step1 > 0 && $abs_feed_to > 0 } {

    set dist [expr $abs_step1 + $abs_rapid_to]

    while { $dist < $abs_feed_to } {

      VNC_move_linear_zcs CUT [ vnc_get_tool_pos [expr -1*$dist] ] 

	  set prev_dist [expr $dist - $abs_step2]
      VNC_move_linear_zcs RAPID [ vnc_get_tool_pos [expr -1*$prev_dist]] 

      set prev_dist $dist
      set dist [expr $dist + $abs_step1 ]
    }
  }

  VNC_move_linear_zcs CUT \
       X $mom_cycle_feed_to_pos(0) \
       Y $mom_cycle_feed_to_pos(1) \
       Z $mom_cycle_feed_to_pos(2) 

  VNC_move_linear_zcs RAPID \
       X $mom_cycle_retract_to_pos(0) \
       Y $mom_cycle_retract_to_pos(1) \
       Z $mom_cycle_retract_to_pos(2) 
}

proc MOM_msys {} {

# set the current ZCS which is the coordinate system that corresponds
# to MCS and is used during simulation. Note: don't position ZCS 
# on a K-component that moves or rotate since ZCS most of
# time has to be stationary (except a table mounted on linear axis
# and the control expect the coordinates in the moving ZCS coordinate system).
# The ZCS should represent the working coordinate system as defined
# on machine tool.
#
#>>>

  # set the current ZCS which is the coordinate system that corresponds
  # to MCS and is used during simulation.
    global mom_last_pos
    global mom_msys_matrix
    global mom_msys_origin
    global gkw_init_msys_matrix 
    global gkw_init_msys_origin
    global gkw_rel_msys_origin
    global gkw_work_plane
    global gkw_work_plane_index
    global gkw_prev_work_plane_index
    global gkw_new_tool
    global sim_prev_tool_name
    global gkw_label_call_number
    global B C
    global gkw_B_work_plane gkw_C_work_plane

# fill matrix with msys   

    MSYS2MAT Mtarget mom_msys_matrix

# fill matrix with reference msys   

    MSYS2MAT Mref gkw_init_msys_matrix

# Invert Mref

    MOM_SIM_MatInvert invMref Mref

# Multiply Matrices M1 is Mtarget relative to Mref
    
    MOM_SIM_MatMultMat M1 invMref Mtarget
    
# Calculate origin of mom_msys_origin i.r.t. gkw_init_msys_origin

    map_point_to_matrix gkw_rel_msys_origin gkw_init_msys_origin gkw_init_msys_matrix mom_msys_origin
    
#    MOM_output_literal "Relative MSYS" 
#
#    for { set i 1 } {$i < 4} {incr i} \
#    {
#      for { set j 1 } {$j < 4} {incr j} \
#      {
#        MOM_output_literal [format "%-1.6f\t,%1i,\t%1i" $M1($j,$i) $j $i]
#      }
#    }
#    MOM_output_literal [format "%-1.6f\t,X" $gkw_rel_msys_origin(0)]
#    MOM_output_literal [format "%-1.6f\t,Y" $gkw_rel_msys_origin(1)]
#    MOM_output_literal [format "%-1.6f\t,Z" $gkw_rel_msys_origin(2)]

# Store info in matrix
    set k 0
    for { set i 1 } {$i < 4} {incr i} \
    {
      for { set j 1 } {$j < 4} {incr j} \
      {
        set gkw_rel_msys($k) $M1($j,$i)
        incr k
      }
    }
    for { set i 0 } {$i < 3} {incr i} \
    {
      set gkw_rel_msys([expr $i+9]) $gkw_rel_msys_origin($i)
    }
     

# add matrix to existing matrix

    set gkw_work_plane_index [add_matrix gkw_rel_msys gkw_work_plane]
    if {![info exists gkw_prev_work_plane_index]} \
    {
      set gkw_prev_work_plane_index -1
    }

    if {$gkw_prev_work_plane_index != $gkw_work_plane_index || $gkw_new_tool == 1} \
    {
      set gkw_label_call_number $gkw_work_plane_index
      MOM_do_template label_call_output
      VNC_rapid_home
    }
    set gkw_prev_work_plane_index $gkw_work_plane_index
    set gkw_new_tool 0
#    set prev_junction [SIM_ask_current_ref_junction]
#    MOM_output_literal $prev_junction
#    VNC_set_ref_jct "undefined"
#    VNC_rapid_home
#    VNC_set_ref_jct $prev_junction

    MSYS2MAT square_msys_matrix gkw_rel_msys
#    calc_angles_from_mat square_msys_matrix alpha beta gamma
    calculate_cycle_19_angles gkw_rel_msys alpha beta gamma

#  MOM_output_literal [format "G54 X%-5.3f Y%-5.3f Z%-5.3f" $gkw_work_plane($i,9) $gkw_work_plane($i,10) $gkw_work_plane($i,11)]
#  MOM_output_literal [format "Cycle 19 A%-5.3f B%-5.3f C%-5.3f" $alpha $beta $gamma]

    VNC_msys "ROTARY_TABLE"

    MTX3_z gkw_rel_msys vecZ

#  MOM_output_literal "I $vecZ(0) J$vecZ(1) K$vecZ(2)"

    vecZ_2_BC vecZ gkw_B_work_plane gkw_C_work_plane

    VNC_rapid_bc
    
#  MOM_output_literal "B$B C$C"

# Force MOM with new Z value so next rapid motion is XY and then Z

    set mom_last_pos(0) -9999.
    set mom_last_pos(1) -9999.
    set mom_last_pos(2) 9999.
    MOM_reload_variable -a mom_last_pos
}

proc MOM_operator_message {} {

     global mom_operator_message

     MOM_output_literal "*$mom_operator_message"
     MOM_start_of_group

  VNC_send_message "OPERATOR:$mom_operator_message"
}
proc LOAD_5AXIS_KINEMATICS {} {

# Kinematics for the DECKEL MAHO DMC 80U Head Table 5 Axis machine
#
# Primery Axis - C Axis 
#    This is the horizontal table with rotary axis Z 
#
# Secondary Axis - C Axis
#    This is the table with the rotary Axis inclined at approximately 45deg
#    to the Z axis of MCS
#
#

  global mom_kin_x_axis_limit
  global mom_kin_y_axis_limit
  global mom_kin_z_axis_limit

  global mom_kin_machine_resolution
  global mom_kin_machine_type
  global mom_kin_4th_axis_plane
  global mom_kin_4th_axis_vector
  global mom_kin_4th_axis_angles
  global mom_kin_4th_axis_min_incr 
  global mom_kin_4th_axis_min_limit 
  global mom_kin_4th_axis_max_limit 
  global mom_kin_4th_axis_direction 
  global mom_kin_4th_axis_rotation 
  global mom_kin_4th_axis_zero 
  global mom_kin_4th_axis_center_offset 
  global mom_kin_pivot_gauge_offset 
#  global mom_kin_4th_to_tracking_point
  global mom_kin_gauge_to_pivot

  global mom_kin_5th_axis_plane 
  global mom_kin_5th_axis_vector 
  global mom_kin_5th_axis_angles 
  global mom_kin_5th_axis_min_incr 
  global mom_kin_5th_axis_min_limit 
  global mom_kin_5th_axis_max_limit 
  global mom_kin_5th_axis_direction 
  global mom_kin_5th_axis_rotation 
  global mom_kin_5th_axis_inclination_to_4th 
  global mom_kin_5th_axis_zero 
  global mom_kin_5th_axis_center_offset

  global mom_kin_linearization_tol 
  global mom_kin_linearization_flag 
  global mom_kin_independent_head 
  global mom_kin_nurbs_output_type 
  global mom_kin_output_unit 

  global mom_kin_4th_axis_center_offset_x
  global mom_kin_4th_axis_center_offset_y
  global mom_kin_4th_axis_center_offset_z
  global mom_kin_4th_axis_angles_a1
  global mom_kin_4th_axis_angles_a2
  global mom_kin_4th_axis_vector_i
  global mom_kin_4th_axis_vector_j
  global mom_kin_4th_axis_vector_k
  global mom_kin_track_center_of_ball
  global mom_kin_read_ahead_next_motion
  
  global leader_4th leader_5th
  global tcl_platform

  if {$tcl_platform(os) == "HP-UX"} {
      set plat "hpp"
      set suffix "sl"
  } elseif {$tcl_platform(os) == "windows" || \
            $tcl_platform(os) == "Windows NT"} {
      set plat "wnt"
      set suffix "dll"
  } else {
      set plat "sun"
      set suffix "so"
  }

  set uf_library "ugpadvkins.$suffix"

#
#
# NOTE: the kinematics parameters commented with the prefix #SIM
#       are set automatically by the SIM_set_post_kinematics function
#       called later in this procedure.
#


  set mom_kin_read_ahead_next_motion             TRUE
  set mom_kin_machine_resolution                 0.000001
  set mom_kin_machine_type                       5_axis_dual_table
  set mom_kin_track_center_of_ball               FALSE

################################################################
# NOTE:
# the variables marked with #SIM are later set based on the
# machine tool model. They are also set here in case the machine
# model doesn't exist.
#################################################################
#SIM  
  set mom_kin_4th_axis_plane                     NONE
#SIM  
  set mom_kin_4th_axis_vector(0)                 0.0000
#SIM  
  set mom_kin_4th_axis_vector(1)                 -1.0
#SIM  
  set mom_kin_4th_axis_vector(2)                 1.0
#  set mom_kin_4th_axis_angles(0)                 0.0000
#  set mom_kin_4th_axis_angles(1)                 0.0000
  set mom_kin_4th_axis_min_incr                  0.001
#SIM  
  set mom_kin_4th_axis_min_limit                 -0.0
#SIM  
  set mom_kin_4th_axis_max_limit                 180.0
#  set mom_kin_4th_axis_direction                 MAGNITUDE_DETERMINES_DIRECTION
  set mom_kin_4th_axis_direction                SIGN_DETERMINES_DIRECTION
  set mom_kin_4th_axis_rotation                  reverse
  set mom_kin_4th_axis_zero                      0
#SIM  
  set mom_kin_4th_axis_center_offset(0)          0.0
#SIM  
  set mom_kin_4th_axis_center_offset(1)          0.0
#SIM  
  set mom_kin_4th_axis_center_offset(2)          0.0

#SIM  
  set mom_kin_pivot_gauge_offset                 0.0

# SATYA:-These varaibles removed to check are added to check......

#SIM  
  set mom_kin_gauge_to_pivot(0)                  0.0
#SIM  
  set mom_kin_gauge_to_pivot(1)                  0.0
#SIM  
  set mom_kin_gauge_to_pivot(2)                  0.0

#  set mom_kin_4th_to_tracking_point(0)           -0.023
#  set mom_kin_4th_to_tracking_point(1)           99.983
#  set mom_kin_4th_to_tracking_point(2)           99.990

#SIM  
  set mom_kin_5th_axis_plane                     NONE
#SIM  
  set mom_kin_5th_axis_vector(0)                 0.0000
#SIM  
  set mom_kin_5th_axis_vector(1)                 0
#SIM  
  set mom_kin_5th_axis_vector(2)                 1
#  set mom_kin_5th_axis_angles(0)                 0.0000
#  set mom_kin_5th_axis_angles(1)                 0.0000
  set mom_kin_5th_axis_min_incr                  0.001
#SIM  
  set mom_kin_5th_axis_min_limit                 0.0
#SIM  
  set mom_kin_5th_axis_max_limit                 180.0
#   set mom_kin_5th_axis_direction                 MAGNITUDE_DETERMINES_DIRECTION
  set mom_kin_5th_axis_direction                 SIGN_DETERMINES_DIRECTION
  set mom_kin_5th_axis_rotation                  reverse
#SIM  
  set mom_kin_5th_axis_inclination_to_4th        45.000
  set mom_kin_5th_axis_zero                      0
#SIM  
  set mom_kin_5th_axis_center_offset(0)          0.0

#          Distance Glynne Daniels testmachine C>B
#  set mom_kin_5th_axis_center_offset(1)          333.67
#  set mom_kin_5th_axis_center_offset(2)          -166.835

#          Distance Spec Tool & Die DMU70V C>B
# N/C software number  is 280476 06
# PLC software number  is GR_482.00C
#  set mom_kin_5th_axis_center_offset(1)          309.96
#  set mom_kin_5th_axis_center_offset(2)          -154.98

#          Distance Cobra WaterTech DMU50V C>B
# N/C software number  is 280476 07
# PLC software number  is GR_482.00D
#  set mom_kin_5th_axis_center_offset(1)           310.02
#  set mom_kin_5th_axis_center_offset(2)          -155.01

#          Distance LEGO DMU70V C>B
# N/C software number  is 280476 06
# PLC software number  is GR_482.00D
# set mom_kin_5th_axis_center_offset(1)          309.86
# set mom_kin_5th_axis_center_offset(2)          -154.928
  
#          Distance DMG (LUTON) DMU50V C>B
# N/C software number  is 280476 16
# PLC software number  is GR_482.00H
# set mom_kin_5th_axis_center_offset(1)          309.994
#set mom_kin_5th_axis_center_offset(2)          -154.997
  
#          Distance DMG-2 (LUTON) DMU50V C>B
# N/C software number  is 280476 16
# PLC software number  is GR_482.00H
# set mom_kin_5th_axis_center_offset(1)          310.002
# set mom_kin_5th_axis_center_offset(2)          -155.001

#          Distance DMG-1 (19B VALUE) (MOLDSYSTEMS) DMU50V C>B
# N/C software number  is 280476 16
# PLC software number  is GR_482.00H
# (OFFSET 1 IS DOUBLE +VE VALUE OF OFFSET 2)
#  set mom_kin_5th_axis_center_offset(1)          310.10
#  set mom_kin_5th_axis_center_offset(2)          -155.050


# --- Testing CK's case by Satya ----------------------

#  set mom_kin_5th_axis_center_offset(1)          0.0
#  set mom_kin_5th_axis_center_offset(2)          0.0

#------------------------------------------------------
#  set mom_kin_5th_axis_center_offset(0)          0.0
#  set mom_kin_5th_axis_center_offset(1)          0.0
#  set mom_kin_5th_axis_center_offset(2)          0.0


#
# this offsets are based on the DMU50 graphical model
#SIM  
#  set mom_kin_5th_axis_center_offset(1)          -523.596
#SIM  
#  set mom_kin_5th_axis_center_offset(2)          -187.046

# Values used for DMU 50 eVolution at Mevi
#SIM  
  set mom_kin_5th_axis_center_offset(1)          380.970985777
#SIM  
  set mom_kin_5th_axis_center_offset(2)          -225.970985777


  set mom_kin_linearization_tol                  0.001
  set mom_kin_linearization_flag                 TRUE ; # T ON, F OFF
  set mom_kin_independent_head                   NONE
  set mom_kin_nurbs_output_type                  BSPLINE

#
# Look at that... this is realy confusing
  set leader_4th "C"
  set leader_5th "B"


  #
  # set kinematics based on the machine tool kinematics model
  # SWIVEL_TABLE@ROTARY_JCT represents the MCS used with this machine
  if { [ VNC_is_machine_tool_model_exists ] } {
      VNC_set_post_kinematics SWIVEL_TABLE@ROTARY_JCT X Y Z C B
  }

# The globals below are to enable ufunc to read the kinematics
# value of center_offset_x/y/z is the same with  above 
# mom_kin_4th_axis_center_offset(i), tcl/ufun can not read array


  global mom_kin_5th_axis_center_offset_x
  global mom_kin_5th_axis_center_offset_y
  global mom_kin_5th_axis_center_offset_z
  global mom_kin_5th_axis_angles_a1
  global mom_kin_5th_axis_angles_a2
  global mom_kin_5th_axis_vector_i
  global mom_kin_5th_axis_vector_j
  global mom_kin_5th_axis_vector_k
#  global mom_kin_4th_to_tracking_point_x
#  global mom_kin_4th_to_tracking_point_y
#  global mom_kin_4th_to_tracking_point_z
  global mom_kin_gauge_to_pivot_x
  global mom_kin_gauge_to_pivot_y
  global mom_kin_gauge_to_pivot_z
  global mom_kin_4th_axis_limit_action
  global mom_kin_5th_axis_limit_action
  
  set mom_kin_4th_axis_limit_action             "Retract / Re-Engage"
  set mom_kin_5th_axis_limit_action             "Retract / Re-Engage"
  
  set mom_kin_4th_axis_center_offset_x $mom_kin_4th_axis_center_offset(0)
  set mom_kin_4th_axis_center_offset_y $mom_kin_4th_axis_center_offset(1) 
  set mom_kin_4th_axis_center_offset_z $mom_kin_4th_axis_center_offset(2)

  set mom_kin_5th_axis_center_offset_x $mom_kin_5th_axis_center_offset(0)
  set mom_kin_5th_axis_center_offset_y $mom_kin_5th_axis_center_offset(1) 
  set mom_kin_5th_axis_center_offset_z $mom_kin_5th_axis_center_offset(2)

  set mom_kin_4th_axis_vector_i $mom_kin_4th_axis_vector(0)
  set mom_kin_4th_axis_vector_j $mom_kin_4th_axis_vector(1)
  set mom_kin_4th_axis_vector_k $mom_kin_4th_axis_vector(2)

  set mom_kin_5th_axis_vector_i $mom_kin_5th_axis_vector(0)
  set mom_kin_5th_axis_vector_j $mom_kin_5th_axis_vector(1)
  set mom_kin_5th_axis_vector_k $mom_kin_5th_axis_vector(2)

# SATYA:-These varaibles removed to check are added to check......

#  set mom_kin_gauge_to_pivot_x $mom_kin_gauge_to_pivot(0)
#  set mom_kin_gauge_to_pivot_y $mom_kin_gauge_to_pivot(1)
#  set mom_kin_gauge_to_pivot_z $mom_kin_gauge_to_pivot(2)

#  set mom_kin_4th_to_tracking_point_x $mom_kin_4th_to_tracking_point(0)
#  set mom_kin_4th_to_tracking_point_y $mom_kin_4th_to_tracking_point(1)
#  set mom_kin_4th_to_tracking_point_z $mom_kin_4th_to_tracking_point(2)

#  set mom_kin_4th_axis_angles_a1 $mom_kin_4th_axis_angles(0)
#  set mom_kin_4th_axis_angles_a2 $mom_kin_4th_axis_angles(1)

#  set mom_kin_5th_axis_angles_a1 $mom_kin_5th_axis_angles(0)
#  set mom_kin_5th_axis_angles_a2 $mom_kin_5th_axis_angles(1)


  switch $mom_kin_4th_axis_plane {
    "XY" -
    "YX"
         {
           set mom_kin_4th_axis_vector_i                  0.0
           set mom_kin_4th_axis_vector_j                  0.0
           set mom_kin_4th_axis_vector_k                  1.0
         }
    "YZ" -
    "ZY"
         {
           set mom_kin_4th_axis_vector_i                  1.0
           set mom_kin_4th_axis_vector_j                  0.0
           set mom_kin_4th_axis_vector_k                  0.0
         }
    "ZX" -
    "XZ"
         {
           set mom_kin_4th_axis_vector_i                  0.0
           set mom_kin_4th_axis_vector_j                  1.0
           set mom_kin_4th_axis_vector_k                  0.0
         }
  default
         {
           if {[info exists mom_kin_4th_axis_vector]} {
               set mom_kin_4th_axis_vector_i $mom_kin_4th_axis_vector(0)
               set mom_kin_4th_axis_vector_j $mom_kin_4th_axis_vector(1)
               set mom_kin_4th_axis_vector_k $mom_kin_4th_axis_vector(2)
           } elseif {[info exists mom_kin_4th_axis_angles]} {
               set mom_kin_4th_axis_angles_a1 $mom_kin_4th_axis_angles(0)
               set mom_kin_4th_axis_angles_a2 $mom_kin_4th_axis_angles(1)
           } else {
               set mom_kin_4th_axis_vector_i                  0.0
               set mom_kin_4th_axis_vector_j                  0.0
               set mom_kin_4th_axis_vector_k                  0.0
               set mom_kin_4th_axis_angles_a1                 0.0
               set mom_kin_4th_axis_angles_a2                 0.0
               MOM_output_to_listing_device \
                   "ERROR: Machine's 4th axis Kinematics are INCONSISTANT"
           }
         }
  }

  switch $mom_kin_5th_axis_plane {
    "XY" -
    "YX"
         {
           set mom_kin_5th_axis_vector_i                  0.0
           set mom_kin_5th_axis_vector_j                  0.0
           set mom_kin_5th_axis_vector_k                  1.0
         }
    "YZ" -
    "ZY"
         {
           set mom_kin_5th_axis_vector_i                  1.0
           set mom_kin_5th_axis_vector_j                  0.0
           set mom_kin_5th_axis_vector_k                  0.0
         }
    "ZX" -
    "XZ"
         {
           set mom_kin_5th_axis_vector_i                  0.0
           set mom_kin_5th_axis_vector_j                  1.0
           set mom_kin_5th_axis_vector_k                  0.0
         }
  default
         {
           if {[info exists mom_kin_5th_axis_vector]} {
               set mom_kin_5th_axis_vector_i $mom_kin_5th_axis_vector(0)
               set mom_kin_5th_axis_vector_j $mom_kin_5th_axis_vector(1)
               set mom_kin_5th_axis_vector_k $mom_kin_5th_axis_vector(2)
           } elseif {[info exists mom_kin_5th_axis_angles]} {
               set mom_kin_5th_axis_angles_a1 $mom_kin_5th_axis_angles(0)
               set mom_kin_5th_axis_angles_a2 $mom_kin_5th_axis_angles(1)
           } else {
               set mom_kin_5th_axis_vector_i                  0.0
               set mom_kin_5th_axis_vector_j                  0.0
               set mom_kin_5th_axis_vector_k                  0.0
               set mom_kin_5th_axis_angles_a1                 0.0
               set mom_kin_5th_axis_angles_a2                 0.0
               MOM_output_to_listing_device \
                   "ERROR: Machine's 5th axis Kinematics are INCONSISTANT"
           }
         }
  }

  MOM_run_user_function [MOM_ask_env_var UGII_CAM_AUXILIARY_DIR]$uf_library ufusr
  MOM_load_kinematics
}

# MSYS2MAT
# procedure to convert mom_msys_matrix(linear) to square matrix
#
# Input                     linear_matrix(0) 0-8
#
# Output                    square_matrix(1,1) (1,1)-(3,3)
#
# Arguments                 square_matrix, linear_matrix
#
# Result                    square_matrix
#
proc MSYS2MAT {square_matrix linear_matrix} \
{
  upvar 1 $linear_matrix LIN_MAT
  upvar 1 $square_matrix RESULT

  set length [array size LIN_MAT]
  if { $length < 9 } {return 0}
  
  set k 0
  for { set i 1 } {$i < 4} {incr i} \
  {
    for { set j 1 } {$j < 4} {incr j} \
    {
      set RESULT($j,$i) $LIN_MAT($k)
      incr k
    }
  }
}
# map_point_to_matrix
# procedure to calculate position of origin_target i.r.t. origin_ref and matrix_ref
#
# Input                     origin_ref(0) 0-2
#                           matrix_ref(0) 0-8
#                           origin_target(0) 0-2
#
# Output                    result(0) 0-2
#
# Arguments                 result, origin_ref, matrix_ref, origin_target
#
# Result                    result
#

proc map_point_to_matrix {result origin_ref matrix_ref origin_target} \
{
  upvar 1 $result RESULT
  upvar 1 $origin_ref Oref
  upvar 1 $matrix_ref Mref
  upvar 1 $origin_target Otarg
  
  set X [expr $Otarg(0)-$Oref(0)]
  set Y [expr $Otarg(1)-$Oref(1)]
  set Z [expr $Otarg(2)-$Oref(2)]
  
  set RESULT(0) [expr $X*$Mref(0)+$Y*$Mref(1)+$Z*$Mref(2)]
  set RESULT(1) [expr $X*$Mref(3)+$Y*$Mref(4)+$Z*$Mref(5)]
  set RESULT(2) [expr $X*$Mref(6)+$Y*$Mref(7)+$Z*$Mref(8)]
}  
  
# add_matrix
# procedure to add matrix_add to matrix_ref
#
# Input                     matrix_ref(200,11) 200-...,0-11
#                           matrix_comp(11) 0-11
#
# Output                    return value index number where matrix_add is stored
#                           this can be an existing index
#
# Arguments                 matrix_add, matrix_ref
#
# Result                    0 (error) or index number
#

proc add_matrix {matrix_add matrix_ref} \
{
  upvar 1 $matrix_add ADD
  upvar 1 $matrix_ref REF
  
  if {[array size ADD] != 12} {return 0}
  
  set array_size [expr [array size REF]/12]
  set ret_val [expr $array_size+200]
  
  for {set i 0} {$i < $array_size} {incr i} \
  {
    set k [expr $i + 200]
    set found 0
    for {set j 0} {$j < 12} {incr j} \
    {
      if { $REF($k,$j) == $ADD($j) } {incr found}
    }
    if { $found == 12 } \
    {
      set ret_val $k
      set i $array_size
    }
  }
  if {$ret_val == [expr $array_size+200]} \
  {
    # copy matrix_add to matrix_ref
    for {set i 0} {$i < 12} {incr i} \
    {
      set REF($ret_val,$i) $ADD($i)
    }
  }
  return $ret_val
}

# calc_angles_from_mat
#  Description:
#    Retrieves three rotation angles vector from a given 3x3 matrix.
#     
#  Parameter:
#    Input    mat   -- 3x3 matrix
#    Output   alpha -- rotation angle round the x-axes  (degrees)
#             beta  -- rotation angle round the y-axes  (degrees)
#             gamma -- rotation angle round the z-axes  (degrees)
#
#
#  Function value:
#      0 -- OK
#      1 -- invalid matrix
#     
#  History:
#
#     JM    9.07.1996 Written in C
#     GKW   1.10.2002 Converted to TCL
#     

proc calc_angles_from_mat {mat alpha beta gamma} \
{
  upvar 1 $mat M
  upvar 1 $alpha a
  upvar 1 $beta b
  upvar 1 $gamma c
  
#  initialize the return angles to 0
  set EPS 1.e-10
  set pi [ expr asin(1) * 2]
  set a 0.
  set b 0.
  set c 0.

# we have to distinguish between three cases

 if { [expr abs($M(1,3)-1.)] > $EPS && [expr abs($M(1,3)+1.)] > $EPS } \
 {

#        case 1: b is not equal to  90 + n*180 degrees  and
#	              not equal to -90 - n*180 degrees ( n>=0 ).  
#
#
#        Check if matrix can be a rotation matrix, otherwise we
#	 return with error code

    if { [expr abs($M(2,3))] < $EPS && [expr abs($M(3,3))] < $EPS } { return 1}
    if { [expr abs($M(1,1))] < $EPS && [expr abs($M(1,2))] < $EPS } { return 1}

#        Calculate the angles in degrees

   set a [expr atan2(-$M(2,3),$M(3,3)) / $pi * 180.]
   set b [expr asin($M(1,3)) / $pi * 180.]
   set c [expr atan2(-$M(1,2),$M(1,1)) / $pi * 180.]

  } \
  else \
  {

#      For the last two cases the angle a can always be 0

    set a 0.
    if { [expr abs($M(1,3)-1.)] < $EPS } \
    {

#           case 2: b is equal to  90 + n*180 degrees (n>=0)

      if { [expr abs($M(3,2))] < $EPS && [expr abs($M(3,1))] < $EPS } { return 1 }
      set c [expr atan2($M(3,2),-$M(3,1)) / $pi * 180.]
    } \
    else \
    {

#           case 3: b is equal to  90 - n*180 degrees (n>=0)

      if { [expr abs($M(2,1))] < $EPS && [expr abs($M(2,2))] < $EPS } { return 1 }
      set c [expr atan2($M(2,1),-$M(2,2)) / $pi * 180.]
    }

    set b [expr asin($M(1,3)) / $pi * 180.]

  }


#   Try to avoid rounding errors for some frequent cases 

    if { [expr abs($a)] < $EPS } {set a 0.}
    if { [expr abs($b)] < $EPS } {set b 0.}
    if { [expr abs($c)] < $EPS } {set c 0.}

    if { [expr abs($a-90.)] < $EPS } {set a 90.}
    if { [expr abs($b-90.)] < $EPS } {set b 90.}
    if { [expr abs($c-90.)] < $EPS } {set c 90.}

    if { [expr abs($a+90.)] < $EPS } {set a -90.}
    if { [expr abs($b+90.)] < $EPS } {set b -90.}
    if { [expr abs($c+90.)] < $EPS } {set c -90.}

    if { [expr abs($a-180.)] < $EPS } {set a 180.}
    if { [expr abs($b-180.)] < $EPS } {set b 180.}
    if { [expr abs($c-180.)] < $EPS } {set c 180.}

    if { [expr abs($a+180.)] < $EPS } {set a -180.}
    if { [expr abs($b+180.)] < $EPS } {set b -180.}
    if { [expr abs($c+180.)] < $EPS } {set c -180.}

    return 0

}

# vecZ_2_BC 
#  Transformations used for the Maho-Deckel DMU50V
#  this function is the main routine transforming the the workpostition
#  to coordinates allowing to position the machine 
#
#
# Input paramters:
#               vecZ(0): 0-2 i,j,k of Z-axis
###		Alpha: Angle in degrees, measured in the X-Y plane( alpha)
###		Beta: Angle, in degrees, measured from the X-Y plane (Beta)
#
# Output parameters:
#		B  : B-angle in degrees
#		C  : C-angle in degrees
#		e14: B-angle in degrees
#		e15: C-angle in degrees

proc vecZ_2_BC {vecZ B C} \
{
  upvar 1 $vecZ VEC
  upvar 1 $B e14
  upvar 1 $C e15

	set pi [ expr asin(1) * 2]

  ijk_2_alpha_beta $VEC(0) $VEC(1) $VEC(2) Alpha Beta


#	/* Calculate intermediate values */
	
	set e20 [expr cos($pi*$Beta/180)*cos($pi*$Alpha/180)]
	set e21 [expr cos($pi*$Beta/180)*sin($pi*$Alpha/180)]
	set e22 [expr sin($pi*$Beta/180)]
	set e14 [expr (acos(2*$e22-1))*180/$pi]
	set e23 [expr sqrt($e20*$e20+$e21*$e21+$e22*$e22)]


	if {$e20 > 0.0} \
	{
		if {$e21==0.0} \
		{
			set e15 [expr asin(($e22-$e23)/$e20)*180/$pi]
		} \
		else \
		{
			set e15 [expr asin(($e22-$e23)/sqrt($e20*$e20+$e21*$e21))-atan($e21/$e20)];
			set e15 [expr $e15*180/$pi];
		}
	} \
	elseif {$e20<0} \
	{
		if {$e21==0} \
		{
			set e15 [expr asin(($e22-$e23)/abs($e20))*180/$pi-180];
		} \
		else \
		{
			set e15 [expr asin(($e22-$e23)/sqrt($e20*$e20+$e21*$e21))-atan($e21/$e20)];
			set e15 [expr $e15*180/$pi-180];
		}
	} \
	elseif {$e21>0} \
	{
		set e15 [expr asin(($e22-$e23)/abs($e21))*180/$pi];
	} \
	elseif {$e21<0} \
	{
		set e15 [expr asin(($e22-$e23)/abs($e21))*180/$pi+90];
	} \
	else \
	{
		set e15 0;
	}
		
	set e15 [expr -1*$e15];

	if {$e15<0} \
	{
		set e15 [expr 360+$e15];	# Absolute angle of C-axis is defined. */
	}
	return 0;
}

# get_matrix
# procedure to get lin matrix_get out of matrix_ref with index
#
# Input                     matrix_ref(200,11) 200-...,0-11
#                           index
#                           matrix_get(11) 0-11
#
# Output                    matrix_get(11) 0-11
#
# Arguments                 matrix_get, matrix_ref, index
#
# Result                    matrix_get
#

proc get_matrix {matrix_get matrix_ref index} \
{
  upvar 1 $matrix_get GET
  upvar 1 $matrix_ref REF
  
  for {set i 0} {$i < 12} {incr i} \
  {
    set GET($i) $REF($index,$i)
  }
}

#	ijk_2_alpha_beta
# convert vector into alpha and beta angles
#	Angle in degrees, measured in the X-Y plane( alpha)
#	Angle, in degrees, measured from the X-Y plane (Beta)
#
# Input       i : i-component vector
#             j : j-component vector
#             k : k-component vector
#
# Output      a : alpha angle in degrees, measured in the X-Y plane
#             b : beta angle, in degrees, measured from the X-Y plane 
	

proc ijk_2_alpha_beta { i j k a b } \
{
	upvar 1 $a a1
	upvar 1 $b b1

	set pi [ expr asin(1) * 2]
	
	set a1 [expr (atan2($j,$i))*180/$pi] ;#	/* Alpha value */
	if {$a1 < 0} {set a1 [expr 360 + $a1]}

	if { $i == 0 && $j == 0} { set a1 0}; # /* Clarify undefined status */
	
	set b1 [expr 90-(acos($k)*180/$pi)];#			/* Beta value */

#	if(fabs(beta)<0.001)			/* Avoid un-necessary errors */
#		beta=0.0;

	if { $b1 < -0.0001} {error "Impossible workplane"} 
	if { $b1 < 0.0} {set b1 0.0} 

	return 0;
}

proc ask_roll_pitch_yaw {mat a_roll a_pitch a_yaw} \
{
#/* 
#*   Returns the roll, pitch and yaw angles of rotation about a fixed frame, 
#*   given the 3X3 matrix that describes the rotation.
#* 
#*   The formulas for the rotation angles are from the book: "Introduction
#*   to Robotics, Mechanics and Control" by John Craig, page 41.  Remember that
#*   UG rotation matrices are the transpose of the rotation matrices found in
#*   the book.
#*/
    upvar $mat rot_mat
    upvar $a_roll roll
    upvar $a_pitch pitch
    upvar $a_yaw yaw
    
    set pi [ expr asin(1) * 2]
    set cos_pitch_sq [expr ($rot_mat(0)*$rot_mat(0)) + ($rot_mat(3)*$rot_mat(3))];
#/*
#*   Check for special case where pitch == +90 degrees or pitch == -90 degrees
#*/
    if {[EQ_is_zero $cos_pitch_sq] } \
        {
        set cos_pitch  0.0;
        set cos_yaw 1.0;
        set sin_yaw 0.0;
        set cos_roll $rot_mat(4);

        if {$rot_mat(6) < 0.0} \
            {
#/*
#*           Special case where pitch == +90 degress
#*/
            set sin_pitch 1.0;
            set sin_roll $rot_mat(1);
            } \
        else \
            {
#/*
#*           Special case where pitch == +90 degress
#*/
            set sin_pitch -1.0;
            set sin_roll [expr -($rot_mat(1))];
            }

        validate_sine_cosine cos_roll sin_roll;
        } \
    else \
        {
#/*
#*       General case where cos_pitch != 0.0
#*/
        set cos_pitch [expr sqrt($cos_pitch_sq)];
        set sin_pitch [expr -($rot_mat(6))];

        set cos_roll [expr $rot_mat(8)/$cos_pitch];
        set sin_roll [expr $rot_mat(7)/$cos_pitch];

        set cos_yaw [expr $rot_mat(0)/$cos_pitch];
        set sin_yaw [expr $rot_mat(3)/$cos_pitch];

        validate_sine_cosine cos_pitch  sin_pitch;
        validate_sine_cosine cos_roll   sin_roll;
        validate_sine_cosine cos_yaw    sin_yaw;
        }

    set roll  [expr atan2($sin_roll,$cos_roll) / $pi * 180];
    set pitch [expr atan2($sin_pitch,$cos_pitch) / $pi * 180];
    set yaw   [expr atan2($sin_yaw,$cos_yaw) / $pi * 180];
}


proc validate_sine_cosine {c_cosine c_sine} \
{
    upvar $c_cosine cosine
    upvar $c_sine sine
    if {[EQ_is_ge [expr abs($cosine)] 1.0] || [EQ_is_zero $sine] } \
        {
        set sine 0.0;
        set cosine [SIGN $cosine];
        } \
    elseif {[EQ_is_ge [expr abs($sine)] 1.0] || [EQ_is_zero $cosine] } \
        {
        set sine [SIGN $sine];
        set cosine 0.0;
        } \
    else \
        {
        set temp [expr $sine*$sine + $cosine*$cosine];

        if {![EQ_is_equal $temp 1.0]} \
            {
            if {[expr abs($cosine)] > [expr abs($sine)]} \
                {
                set sine [SIGN [expr $sine*sqrt(1.0 - $cosine*$cosine)]];
                } \
            else \
                {
                set cosine [SIGN [expr $cosine*sqrt(1.0 - $sine*$sine)]];
                }
            }
        }
}

proc SIGN {cosine} \
{
  if {$cosine > 0.0} \
  {
    return 1.0
  } \
  else \
  {
    return -1.0
  }
}
# -------------------------------------------------------------------------------
# recalculate point in relation to new matrix
# -------------------------------------------------------------------------------

proc reposition_point {mat ipos opos} \
{
  upvar $mat matr
  upvar $opos out_pos
  upvar $ipos in_pos

	set pos(0) [expr $in_pos(0)*$matr(0)+$in_pos(1)*$matr(1)+$in_pos(2)*$matr(2)]
	set pos(1) [expr $in_pos(0)*$matr(3)+$in_pos(1)*$matr(4)+$in_pos(2)*$matr(5)]
	set pos(2) [expr $in_pos(0)*$matr(6)+$in_pos(1)*$matr(7)+$in_pos(2)*$matr(8)]
  
  set out_pos(0) $pos(0)
  set out_pos(1) $pos(1)
  set out_pos(2) $pos(2)

	return 0;
}

#/*	==============================================================	*/
#/*	Create matrix defining rotation around a line through the origin
#	and with direction cosinus dir[0],dir[1],dir[2].
#	rotation angle in degrees.  */
#/*	===============================================================	*/

proc MTX3_rot_line { angle direction matrix } \
{
  upvar $direction dir
  upvar $matrix matr

  set pi [expr asin(1) * 2]
	set hk [expr $angle*$pi/180]
	set a $dir(0)
	set b $dir(1)
	set c $dir(2)

	set matr(0) [expr $a*$a*(1-cos($hk))+cos($hk)]
  set	matr(1) [expr $a*$b*(1-cos($hk))-$c*sin($hk)]
	set matr(2) [expr $a*$c*(1-cos($hk))+$b*sin($hk)]
	set matr(3) [expr $a*$b*(1-cos($hk))+$c*sin($hk)]
	set matr(4) [expr $b*$b*(1-cos($hk))+cos($hk)]
	set matr(5) [expr $b*$c*(1-cos($hk))-$a*sin($hk)]
	set matr(6) [expr $a*$c*(1-cos($hk))-$b*sin($hk)]
	set matr(7) [expr $b*$c*(1-cos($hk))+$a*sin($hk)]
	set matr(8) [expr $c*$c*(1-cos($hk))+cos($hk)]
		
	return 0
}



proc calculate_cycle_19_angles {mtx1 alpha1 beta1 gamma1} \
{

# Procedure to find the Cycle 19 angles for a Heidenhain controller
#
# Input                 mtx1(8) : 3x3 matrix in linear array (format == mom_msys_matrix)
#                                 mtx1(0) IX
#                                 mtx1(1) JX
#                                 mtx1(2) KX
#                                 mtx1(3) IY
#                                 mtx1(4) JY
#                                 mtx1(5) KY
#                                 mtx1(6) IZ
#                                 mtx1(7) JZ
#                                 mtx1(8) KZ
# Output                alpha1  : Alpha
#                       beta1   : Beta
#                       gamma1  : Gamma

  upvar $mtx1 mtx
  upvar $alpha1 alpha
  upvar $beta1 beta
  upvar $gamma1 gamma
  global mom_system_tolerance

  set pi [expr asin(1) * 2]

# The order in which the Heidenhain controller calculates the rotations is: Alpha, Beta and Gamma
# To find these angles from the matrix, the proces has to be reversed.
# Therefore we start with Gamma

# Calculate rotation around the Z-axis so that the X-axis rotates into the XZ-plane 

  set gamma [expr atan2($mtx(1),$mtx(0))*180/$pi]

# Calculate rotation matrix for found angle and store this in "rmtx"
  
  set dir(0) 0.0
  set dir(1) 0.0
  set dir(2) 1.0

  MTX3_rot_line [expr $gamma] dir rmtx

# Multiply the rotation matrix with the input matrix the result is stored in result1_mtx 

  MTX3_multiply rmtx mtx result1_mtx

#for {set i 0} {$i < 9} {incr i} \
#{
#  MOM_output_literal [format "initial_mtx($i) = %2.5f" $mtx($i)]
#}

#for {set i 0} {$i < 9} {incr i} \
#{
#  MOM_output_literal [format "result1_mtx($i) = %2.5f" $result1_mtx($i)]
#}

# Calculate rotation around the Y-axis so that the X-axis lines up with X 

  set beta [expr atan2($result1_mtx(2)*-1,$result1_mtx(0))*180/$pi]

# Calculate rotation matrix for found angle and store this in "rmtx"

  set dir(0) 0.0
  set dir(1) 1.0
  set dir(2) 0.0

  MTX3_rot_line [expr $beta] dir rmtx

# Multiply the rotation matrix with the input matrix (previous result) the result is stored in result2_mtx 

  MTX3_multiply rmtx result1_mtx result2_mtx

#  for {set i 0} {$i < 9} {incr i} \
#  {
#    MOM_output_literal [format "result2_mtx($i) = %2.5f" $result2_mtx($i)]
#  }

# Calculate rotation around the X-axis so that the Y-axis lines up with Y 

  set alpha [expr atan2($result2_mtx(5),$result2_mtx(4))*180/$pi]

#   Try to avoid rounding errors for some frequent cases 

  if { [expr abs($alpha)] < $mom_system_tolerance } {set alpha 0.}
  if { [expr abs($beta)] < $mom_system_tolerance } {set beta 0.}
  if { [expr abs($gamma)] < $mom_system_tolerance } {set gamma 0.}

  if { [expr abs($alpha-90.)] < $mom_system_tolerance } {set alpha 90.}
  if { [expr abs($beta-90.)] < $mom_system_tolerance } {set beta 90.}
  if { [expr abs($gamma-90.)] < $mom_system_tolerance } {set gamma 90.}

  if { [expr abs($alpha+90.)] < $mom_system_tolerance } {set alpha -90.}
  if { [expr abs($beta+90.)] < $mom_system_tolerance } {set beta -90.}
  if { [expr abs($gamma+90.)] < $mom_system_tolerance } {set gamma -90.}

  if { [expr abs($alpha-180.)] < $mom_system_tolerance } {set alpha 180.}
  if { [expr abs($beta-180.)] < $mom_system_tolerance } {set beta 180.}
  if { [expr abs($gamma-180.)] < $mom_system_tolerance } {set gamma 180.}

  if { [expr abs($alpha+180.)] < $mom_system_tolerance } {set alpha -180.}
  if { [expr abs($beta+180.)] < $mom_system_tolerance } {set beta -180.}
  if { [expr abs($gamma+180.)] < $mom_system_tolerance } {set gamma -180.}


# All info is now available
# the code below, does a final rotation of the matrix, as a check to see if the matrix
# ends up with 1,0,0,0,1,0,0,0,1

  set dir(0) 1.0
  set dir(1) 0.0
  set dir(2) 0.0

  MTX3_rot_line [expr $alpha] dir rmtx

  MTX3_multiply rmtx result2_mtx result3_mtx

#  for {set i 0} {$i < 9} {incr i} \
#  {
#    MOM_output_literal [format "result3_mtx($i) = %2.5f" $result3_mtx($i)]
#  }
  set retcode 0
  if  {![EQ_is_equal $result3_mtx(0) 1]} {set retcode 1}
  if  {![EQ_is_equal $result3_mtx(1) 0]} {set retcode 1}
  if  {![EQ_is_equal $result3_mtx(2) 0]} {set retcode 1}
  if  {![EQ_is_equal $result3_mtx(3) 0]} {set retcode 1}
  if  {![EQ_is_equal $result3_mtx(4) 1]} {set retcode 1}
  if  {![EQ_is_equal $result3_mtx(5) 0]} {set retcode 1}
  if  {![EQ_is_equal $result3_mtx(6) 0]} {set retcode 1}
  if  {![EQ_is_equal $result3_mtx(7) 0]} {set retcode 1}
  if  {![EQ_is_equal $result3_mtx(8) 1]} {set retcode 1}
  
  if {$retcode} {MOM_abort "Calculation of Cycle 19 Angles failed"}

}


#-------------------------------------------------------------------------------
# GKW_scan_args
#
#	Modul: 					IS&V
#
#	Task:						scans the arguments and store in global vars
#
#	Arguments:				arguments (e.g. X 4452 Z 762 R -72)
#
#	Results:					
#
# 	Last modification:	01-04-2003
#	
# 	Owner:					Gerrit Koelewijn
#-------------------------------------------------------------------------------

proc GKW_scan_args { args } \
{
	global X_sim_pos Y_sim_pos Z_sim_pos B_sim_pos C_sim_pos

	if {[join $args] == ""} \
	{
		#	invalid argument
		MOM_SIM_ERROR_MSG "No arguments defined in GKW_scan_args --> abort"
		MOM_abort
	}
	
	#	parsing the arguments

	foreach {var value}  [join $args] \
	{
		#	scan arguments and store in vars
		set ${var}_sim_pos $value
  }
}

proc VNC_linear_lintol {args} \
{
# Procedure to that splits the SIM_move_linear_zcs in several motions to accomodate lintol
# This is only for purpose of simulation, and not for NC_output
# it takes only into account the addresses that are in the argument list
# MTCS or ZCS are valid arguments, ZCS is default
# These arguments declare the machine or program coordinate system

  global sim_4th_axis sim_5th_axis
  global gkw_B_work_plane gkw_C_work_plane
  global mom_pos
  global B C
	global X_sim_old_pos Y_sim_old_pos Z_sim_old_pos B_sim_old_pos C_sim_old_pos
  global tollin
  global gkw_tpcm_mode

  set SIM_args ""
  set SIM_csys "zcs"
	foreach address [join $args] \
	{
		#	add only the given addresses to the argument list for SIM_move_linear_zcs
		switch $address {
		  X { append SIM_args {X [expr $X_sim_old_pos+($i./$num_lin)*$inc_X] } }
      Y { append SIM_args {Y [expr $Y_sim_old_pos+($i./$num_lin)*$inc_Y] } }
      Z { append SIM_args {Z [expr $Z_sim_old_pos+($i./$num_lin)*$inc_Z] } }
      B { append SIM_args {$sim_4th_axis [expr $B_sim_old_pos+($i./$num_lin)*$inc_B] } }
      C { append SIM_args {$sim_5th_axis [expr $C_sim_old_pos+($i./$num_lin)*$inc_C] } }
      ZCS { set SIM_csys "zcs" }
      MTCS { set SIM_csys "mtcs" }
      default {MOM_abort "invalid arguments"}
    }
  }
  
  set inc_X [expr $mom_pos(0) - $X_sim_old_pos]
  set inc_Y [expr $mom_pos(1) - $Y_sim_old_pos]
  set inc_Z [expr $mom_pos(2) - $Z_sim_old_pos]
  set inc_B [expr $B + $gkw_B_work_plane - $B_sim_old_pos]
  set inc_C [expr $C + $gkw_C_work_plane - $C_sim_old_pos]
  if { $gkw_tpcm_mode == 128 } {
    if { [expr abs($inc_B)] > [expr abs($inc_C)] } \
    {
      set num_lin [expr int(abs($inc_B)/$tollin)+1]
    } \
    else \
    {
      set num_lin [expr int(abs($inc_C)/$tollin)+1]
    }
  } else {
    set num_lin 1
  }
  
  
#MOM_output_literal "num_lin $num_lin"
  for {set i 1} {$i <= $num_lin} {incr i} \
  {
    eval SIM_move_linear_$SIM_csys $SIM_args
  }
  
  set X_sim_old_pos $mom_pos(0)
  set Y_sim_old_pos $mom_pos(1)
  set Z_sim_old_pos $mom_pos(2)
  set B_sim_old_pos [expr $B + $gkw_B_work_plane]
  set C_sim_old_pos [expr $C + $gkw_C_work_plane]

}

